/*
 * machine_dproc.c
 *
 * Copyright (C) 2001, Neil Bright
 *
 * This source is covered by the GNU GPL.  See the file COPYING.
 */

#include "dproc.h"
#include "dproc_int.h"

MODULE_AUTHOR("Neil Bright <ncb@cc.gatech.edu>");
MODULE_DESCRIPTION("Distributed /proc version");
MODULE_LICENSE("GPL");


#define DRV_NAME	"dproc"
#define DRV_VERSION	"1.0"
#define DRV_RELDATE	"June 3, 2001"

//static char version[] = DRV_NAME " version " DRV_VERSION " (" DRV_RELDATE ")";
static struct proc_dir_entry *dproc_root = NULL;
static rwlock_t hashtable_lock = RW_LOCK_UNLOCKED;
static int next_descriptor = 1;
static LIST_HEAD(dproc_hashtable);

/***********************
 * allocate_descriptor *
 *************************************************************************
 *
 * allocate the next available descriptor
 */
inline int
allocate_descriptor(void)
{
  return next_descriptor++;
}



/*********************
 * unhash_descriptor *
 *************************************************************************
 *
 * find a descriptor @d in @dproc_hashtable
 */
static struct dproc_placeholder *
unhash_descriptor( int d )
{
  struct list_head *i;
  struct dproc_placeholder *p;

  list_for_each(i, &dproc_hashtable) {
    p = list_entry(i, struct dproc_placeholder, hash_list);
    if( d == p->descriptor )
      return p;
  }

  return NULL;
}



/*********************
 * proc_calc_metrics *
 *************************************************************************
 *
 * fixup code stolen from fs/proc/proc_misc.c
 */
static int
proc_calc_metrics(char *page, char **start, off_t off, int count, int *eof,
		  int len)
{
  if (len <= off+count) *eof = 1;
  *start = page + off;
  len -= off;
  if (len>count) len = count;
  if (len<0) len = 0;
  return len;
}



/*********************
 * proc_calc_metrics *
 *************************************************************************
 *
 * seems this isn't properly exported from proc/fs/generic.c
 */
int
proc_match(int len, const char *name, struct proc_dir_entry * de)
{
  if (!de || !de->low_ino)
    return 0;
  if (de->namelen != len)
    return 0;
  return !memcmp(name, de->name, len);
}



/**************
 * find_child *
 *************************************************************************
 */
static struct proc_dir_entry *
find_child(const char *name, int length, struct proc_dir_entry *parent)
{
  struct proc_dir_entry *i;

  dprintk("%s: looking for [%s]:%d\n", __PRETTY_FUNCTION__, name, length);

  for( i=parent->subdir; i; i = i->next ) {
    dprintk("%s: comparing to [%s]:%d\n", __PRETTY_FUNCTION__, i->name, i->namelen);

    if( proc_match(length, name, i) ) {
      dprintk("%s: found it!\n", __PRETTY_FUNCTION__);
      return i;
    }
  }

  return NULL;
}



/*****************
 * parse_command *
 *************************************************************************
 *
 * valid input looks something like this:
 *
 * <action> <whitespace> <token>
 *
 * <action> = [ a | d ]
 * <whitespace> = <space>+
 * <token> = <printable char>+
 *
 * A <token> is a nodelist entry or a statslist entry
 */
static char
parse_command(char *kbuf, char **nodename, unsigned long *length)
{
  char *tmp;
  int len = *length;

  kbuf[len] = '\0';

  /* remove a \n if present */
  tmp = strchr(kbuf, '\n');
  if(tmp) {
    len--;
    *tmp = '\0';
  }

  /* find start of the token */
  tmp = strrchr(kbuf, ' ');
  if(!tmp)
    return '\0';
  tmp++;
  len -= (tmp - kbuf);

  if(len > DPROC_NAME_MAX )
    return '\0';

  *length = len;
  *nodename = tmp;
  return kbuf[0];
}



/***************
 * _list_write *
 *************************************************************************
 *
 * Worker function for nodelist_write() and statslist_write()
 */
static int
_list_write(const char *buffer, unsigned long length,
	    struct proc_dir_entry *parent,
	    int (*create_proc)(const char *, struct proc_dir_entry*, kdproc_handler_t *, void *),
	    int (*delete_proc)(const char *, int, struct proc_dir_entry*),
	    char *pretty_function )
{
  char name[DPROC_NAME_MAX  + 1];
  char *tok = NULL;
  char c;


  if(length >= DPROC_NAME_MAX)
    return -EINVAL;
  if(copy_from_user(name, buffer, length))
    return -EFAULT;

  /* ensures NULL termination of tok */
  c = parse_command(name, &tok, &length);

  switch(c) {
  case 'a':
    dprintk("%s: ADDING [%s]\n", pretty_function, tok);
    if( !create_proc(tok, parent, NULL, NULL) )
      return -ENOMEM;
    break;

  case 'd':
    dprintk("%s: REMOVING [%s] from [%s]\n", pretty_function, tok, parent->name);
    if( !delete_proc(tok, length, parent) )
      return -ENOMEM;
    break;

  default:
    dprintk("%s: hit unknown or bad action [%c]\n", pretty_function, c );
    return -EINVAL;
  }

  return length;
}



/*******************
 * statslist_write *
 *************************************************************************
 *
 * write method for /proc/cluster/<node>/statslist
 *
 * @file:		unused
 * @buffer:		placeholder (2) to add/remove
 * @count:		length of @buffer
 * @data:		ptr to (4)
 *
 * Format of a write should look like this:
 *
 * [a|d] <statsname>
 *
 * where the 'a' prefix causes a new placeholder to be created and
 * the 'd' prefix causes an existing placeholder to be deleted.
 */

static int
statslist_write(struct file *file, const char *buffer,
		unsigned long count, void *data)
{
  int ret;
  struct proc_dir_entry *parent = (struct proc_dir_entry *)data;
  char *pretty = __PRETTY_FUNCTION__;

  dprintk("%s: got called!\n", pretty);

  ret = _list_write(buffer, count, parent, dproc_create_placeholder,
		    dproc_delete_placeholder, pretty);

  dprintk("%s: returning %ld\n", pretty, count);
  return ret;
}

/*****************
 * generic_write *
 *************************************************************************
 *
 * write method wrapper for /proc/cluster/<node>/<placeholder>
 *
 * @file:	unused
 * @buffer:	pointer to userspace buffer
 * @count:	size of @buffer
 * @data:	ptr to struct dproc_placeholder of interest
 *
 * Format of write for "controls" is:
 * <cpu_update_time> <mem_update_time> <cpu_percent_change> <mem_percent_change>
 *
 */
static int
generic_write(struct file *file, const char *buffer, unsigned long count,
	      void *data)
{
  return _write((int)data, buffer, count, 0);
}


/**********
 * _write *
 *************************************************************************
 *
 * write method for /proc/cluster/<node>/<placeholder>
 *
 * @descriptor:		placeholder to read from
 * @buffer:		pointer to userspace buffer
 * @count:		size of @buffer
 * @flag:		set if @buffer is in kernel space, else zero
 *
 */
static int
_write(int descriptor, const char *buffer, unsigned long count, int flag)
{
  struct dproc_placeholder *d;
  int ret;
  char temp_buf[DPROC_DATA_SIZE];

  dprintk("%s: got called! [%d][%d]\n", __PRETTY_FUNCTION__, descriptor, flag);

  /* don't let placeholders get unhashed out from under me */
  read_lock(&hashtable_lock);

  /* locate placeholder */
  if( !(d = unhash_descriptor(descriptor)) )
    goto err_hash;

  /* lock in write mode to block other user requests */
  write_lock(&(d->lock));

  /* copy apropriately sized data */
  ret = MIN(count, DPROC_DATA_SIZE);

  /* The kernel interface sets this parameter to non-zero.  Since I'm doing
   * a kernel-to-kernel operation, don't use copy_from_user!!!
   */
  if( flag ) {
    memcpy(d->buffer, buffer, ret);
    d->length = ret;
  } else {
    if(copy_from_user(temp_buf, buffer, ret))
      goto err_copy;
    /*Added by sandip. */
    if(d->kdproc_handler != NULL)
	    d->kdproc_handler(temp_buf, ret, d->node_ptr);
  }


  /* unlock & go */
  write_unlock(&(d->lock));
  read_unlock(&(hashtable_lock));
  
	  //(d->push_control_update)(d->buffer,ret);

  dprintk("%s: returning %d\n", __PRETTY_FUNCTION__, ret);
  return ret;

 err_hash:
  printk("dproc:[%s] couldn't unhash_descriptor(%d)\n",
	 __PRETTY_FUNCTION__, descriptor);

  read_unlock(&hashtable_lock);
  return -EBADFD;

 err_copy:
  printk("dproc:[%s] copy_from_user(0x%p, 0x%p, %d) failed!\n",
	 __PRETTY_FUNCTION__, d->buffer, buffer, ret);

  write_unlock(&(d->lock));
  read_unlock(&hashtable_lock);
  return -EFAULT;
}



/*****************
 * generic_read *
 *************************************************************************
 *
 * read method for /proc/cluster/<node>/<placeholder>
 *
 * @page:	location to write data to
 * @start:	where to start at next time around (or @page + @off)
 * @off:	offset into the file to start at this time
 * @count:	size of buffer pointed to by @page
 * @eof:	set to nonzero when have no more data to provide
 * @node:	struct dproc_node of interest
 * @data:	ptr to struct dproc_placeholder of interest
 *
 * return number of bytes stuffed into @page
 *
 */
static int
generic_read(char *page, char **start, off_t off, int count, int *eof,
	     void *data)
{
  struct dproc_placeholder *d;
  int ret=0;

  dprintk("%s: got called!    page=0x%p start=0x%p off=%ld count=%d eof=%d\n",
	  __PRETTY_FUNCTION__, page, start, off, count, *eof);

  /* don't let placeholders get removed out from under me */
  read_lock(&(hashtable_lock));

  /* locate placeholder */
  if( !(d = unhash_descriptor((int)data)) )
    goto err_hash;

  /* lock in read mode to allow user read requests to proceed in parallel */
  read_lock(&(d->lock));

  /* bounds checking */
  if( !d->length || !d->buffer )	/* empty or invalid buffer */
    return count;

  memcpy(page, d->buffer, d->length);
  ret = proc_calc_metrics(page, start, off, count, eof, d->length);

  /* unlock & go */
  read_unlock(&(d->lock));
  read_unlock(&hashtable_lock);

  dprintk("%s: returning %d    page=0x%p start=0x%p off=%ld count=%d eof=%d\n",
	  __PRETTY_FUNCTION__, ret, page, start, off, count, *eof);
  return ret;

 err_hash:
  read_unlock(&hashtable_lock);
  printk("dproc:[%s] couldn't unhash_descriptor(%d)\n",
	 __PRETTY_FUNCTION__, (int)data);
  return -EBADFD;
}



/****************************
 * dproc_create_placeholder *
 *************************************************************************
 *
 * Create a placeholder.  eg. /proc/cluster/<node>/<name>
 *
 * @name:	name of placeholder (2) to create
 * @parent:	pointer to (4)
 *
 * return descriptor or negative errno
 */
static int
dproc_create_placeholder(const char *name, struct proc_dir_entry *parent, 
		kdproc_handler_t *handler, void *node_ptr)
{
  struct dproc_placeholder *node;
  struct proc_dir_entry *tmp;

  dprintk("%s: got called!\n", __PRETTY_FUNCTION__);

  /* allocate and init a placeholder */
  if( !(node = kmalloc(sizeof(struct dproc_placeholder), GFP_KERNEL)) )
    goto err_placeholder;
  
  node->length = 0;
  node->kdproc_handler = handler;
  node->node_ptr = node_ptr;
  node->descriptor = allocate_descriptor();
  rwlock_init(&(node->lock));

  if( !(node->buffer = kmalloc(DPROC_DATA_SIZE, GFP_KERNEL)) )
    goto err_buff;

  dprintk("%s: buffer allocated\n", __PRETTY_FUNCTION__);

  memset(node->buffer, 0, DPROC_DATA_SIZE);

  dprintk("%s: buffer cleared\n", __PRETTY_FUNCTION__);
  strcpy(node->name,name);		/*Added by sandip.. Needed for "controls"*/
	
  /* link in to hash table */
  write_lock(&hashtable_lock);
  dprintk("%s: locked\n", __PRETTY_FUNCTION__);

  list_add(&(node->hash_list), &dproc_hashtable);

  dprintk("%s: linked\n", __PRETTY_FUNCTION__);
  write_unlock(&hashtable_lock);

  dprintk("%s: linked into hashtable\n", __PRETTY_FUNCTION__);

  /* create and initialize proc entry */
  if( !(tmp = create_proc_entry( name, DPROC_PLACEHOLDER_MODE, parent )) )
    goto err_proc;

  tmp->data = (void*)node->descriptor;
  tmp->read_proc = generic_read;
  tmp->write_proc = generic_write;

  dprintk("%s: returning good [%d]\n", __PRETTY_FUNCTION__, node->descriptor);

  return node->descriptor;

 err_proc:
  remove_proc_entry( name, parent );
  /* fall through */

 err_buff:
  kfree(node->buffer);
  /* fall through */

 err_placeholder:
  kfree(node);
  /* fall through */
  
  return -ENOMEM;
}



/****************************
 * dproc_delete_placeholder *
 *************************************************************************
 *
 * delete a placeholder.  eg. /proc/cluster/<node>/<name>
 *
 * @name:	name of placeholder (2) to delete
 * @length:	length of @name
 * @parent:	pointer to (4)
 */
static int
dproc_delete_placeholder(const char *name, int length,
			 struct proc_dir_entry *parent)
{
  struct proc_dir_entry *tmp;
  struct dproc_placeholder *d;

  dprintk("%s: got called!\n", __PRETTY_FUNCTION__);

  /* locate placeholder and extract descriptor */
  if( !(tmp = find_child(name, length, parent)) )
    goto err_child;

  /* don't unhash placeholders out from under other threads */
  dprintk("%s: locking hashtable\n", __PRETTY_FUNCTION__);
  write_lock(&hashtable_lock);

  dprintk("%s: unhash_descriptor\n", __PRETTY_FUNCTION__);
  if( !(d = unhash_descriptor((int)tmp->data)) )
    goto err_hash;

  /* locking in write mode ensures completion of all reads on this placeholder */
  dprintk("%s: locking node\n", __PRETTY_FUNCTION__);
  write_lock(&(d->lock));
  
  /* destroy proc node */
  dprintk("%s: removing entry\n", __PRETTY_FUNCTION__);
  tmp->read_proc = NULL;
  tmp->write_proc = NULL;
  tmp->data = NULL;
  remove_proc_entry(name, parent);

  /*
   * Mark this placeholder invalid, but don't actually remove from hastable.
   * generic_{read|write} must check against a NULL buffer!!
   */
  dprintk("%s: freeing buffer\n", __PRETTY_FUNCTION__);
  kfree(d->buffer);
  d->buffer = NULL;
  d->length = 0;

  dprintk("%s: unlocking\n", __PRETTY_FUNCTION__);
  write_unlock(&(d->lock));
  write_unlock(&hashtable_lock);

  return 0;

 err_child:
  dprintk("%s: Hit err_child!\n", __PRETTY_FUNCTION__);
  return -ENOENT;

 err_hash:
  dprintk("%s: Hit err_hash!\n", __PRETTY_FUNCTION__);
  write_unlock(&hashtable_lock);
  return -EBADFD;
}



/****************
 * _create_node *
 *************************************************************************
 *
 * Create a new node.  eg. /proc/cluster/<name>
 *
 * @name:	name of node (4) to create
 * @parent:	pointer to (6)
 * @ret:	(OUTPUT) set to point to the newly created node (4)
 *
 * returns 0 (and sets @ret), else negative errno
 *
 */
static int
_create_node(const char *name, struct proc_dir_entry *parent,
	     struct proc_dir_entry **ret)
{
  struct proc_dir_entry *node, *stats;

  if( !(node = proc_mkdir(name, parent)) )		// node is ptr to (4)
    goto err_mkdir;
  if( !(stats = create_proc_entry(DPROC_STATSLIST, DPROC_STATSLIST_MODE, node)) )
    goto err_statslist;
  
  stats->data = node;
  stats->write_proc = statslist_write;

  *ret = node;
  return 0;

 err_statslist:
  remove_proc_entry(DPROC_STATSLIST, node);
 
  /* fall through */

 err_mkdir:
  remove_proc_entry(name, parent);
  return -ENOMEM;
}

/*********************
 * dproc_delete_node *
 *************************************************************************
 *
 * Delete an existing node.  eg. /proc/cluster/<name>
 *
 * @name:	name of node (4) to delete
 * @length:	length of @name
 * @parent:	pointer to (6)
 *
 */
static int
dproc_delete_node(const char *name, int length, struct proc_dir_entry *parent)
{
  struct proc_dir_entry *node;		// ptr to (4)

  dprintk("%s: got called!\n", __PRETTY_FUNCTION__);
  
  if( !(node = find_child(name, length, parent)) ) {
    dprintk("%s: find_child([%s] [%d]) didn't\n",
	    __PRETTY_FUNCTION__, name, length);
    return -ENOENT;
  }

  /* remove the STATSLIST (3) from the appropriate node (4) to ensure that
   * all remaining entries in node->subdir are statistics (2)
   */
  dprintk("%s: removing statslist from node [%s]\n",
	  __PRETTY_FUNCTION__, node->name);
  remove_proc_entry(DPROC_STATSLIST, node);

  /* remove any remaining placeholders */
  dprintk("%s: removing placeholders from node [%s]\n",
	  __PRETTY_FUNCTION__, node->name);

  while( node && node->subdir ) {
    dprintk("%s: [%s] [%d]\n", __PRETTY_FUNCTION__, node->subdir->name,
	    node->subdir->namelen);
    dproc_delete_placeholder(node->subdir->name, node->subdir->namelen, node);
  }

  /* remove the node (4) itself */
  dprintk("%s: removing node [%s]!\n", __PRETTY_FUNCTION__, name);
  remove_proc_entry(name, parent);
  return 0;
}



/**************
 * dproc_exit *
 *************************************************************************
 *
 * Module exit routine
 */
static
void __exit
dproc_exit(void)
{
  struct list_head *i;
  struct dproc_placeholder *p;

  dprintk("%s: got called!\n", __PRETTY_FUNCTION__);
//  printk( KERN_INFO "%s exiting\n", version);

  /* remove all dproc entries in procfs */
  remove_proc_entry(DPROC_ROOTDIR, NULL);

  write_lock(&hashtable_lock);

  /* free all of the buffers and dproc_placeholders */

  i = dproc_hashtable.next;
  while( i != dproc_hashtable.next ) {
    p = list_entry(i, struct dproc_placeholder, hash_list);
    dprintk("%s: freeing descriptor %d\n", __PRETTY_FUNCTION__, p->descriptor);
    kfree(p->buffer);
    i = i->next;
    kfree(p);
  }

  write_unlock(&hashtable_lock);
  dprintk("%s: finished cleanup\n", __PRETTY_FUNCTION__);

  /* module is unloaded at this point */
}



/**************
 * dproc_init *
 *************************************************************************
 *
 * Module init routine
 */
static
int __init
dproc_init(void)
{
  struct proc_dir_entry *t; //*nodelist;
  dprintk("%s: got called!\n", __PRETTY_FUNCTION__);

  //printk( KERN_INFO "%s\n", version);

  if( !(dproc_root = proc_mkdir(DPROC_ROOTDIR, NULL)) )
    goto err_out;

  dprintk("dproc_root is at 0x%p\n", dproc_root);
  dprintk("dproc_root is [%s] [%d]\n", dproc_root->name, dproc_root->namelen);

  t = find_child(DPROC_ROOTDIR, strlen(DPROC_ROOTDIR), &proc_root);
  dprintk("found dproc_root at 0x%p\n", t);

  dprintk("%s: finished initialization\n", __PRETTY_FUNCTION__);

  return 0;

  /* fall through */

 err_out:
  dproc_exit();
  return -ENOMEM;
}



/*****************************
 * kdproc_create_placeholder *
 *************************************************************************
 *
 * Kernel interface to create a placeholder.  eg. /proc/cluster/<node>/<name>
 *
 * @name:	name of placeholder (2) to create
 * @parent:	pointer to (4)
 * @handler:	This is mailnly for pushing control updates to the monitor module
 *
 * return descriptor or negative errno
 */
extern int
kdproc_create_placeholder(const char *name, struct proc_dir_entry *parent, 
		kdproc_handler_t *handler, void *node_ptr)
{
	return dproc_create_placeholder(name, parent, handler, node_ptr);
}



/*****************************
 * kdproc_delete_placeholder *
 *************************************************************************
 *
 * Kernel interface to delete a placeholder.  eg. /proc/cluster/<node>/<name>
 *
 * @name:	name of placeholder (2) to delete
 * @length:	length of @name
 * @parent:	pointer to (4)
 *
 * return descriptor or negative errno
 */
extern int
kdproc_delete_placeholder(const char *name, int length,
			  struct proc_dir_entry *parent)
{
  return dproc_delete_placeholder(name, length, parent);
}



/*********************
 * kdproc_create_node *
 *************************************************************************
 *
 * Kernel interface to create a new node.  eg. /proc/cluster/<name>
 *
 * @name:	name of node to create
 * @ret:	(OUTPUT) set to point to the newly created node (4)
 *
 * returns 0 (and sets @ret), else negative errno
 *
 */
extern int
kdproc_create_node(const char *name, struct proc_dir_entry **ret)
{
  return _create_node(name, dproc_root, ret);
}



/*********************
 * kdproc_delete_node *
 *************************************************************************
 *
 * Kernel interface to delete an existing node.  eg. /proc/cluster/<name>
 *
 * @name:	name of node to create
 * @length:	length of @name
 */
extern int
kdproc_delete_node(const char *name, int length)
{
  return dproc_delete_node(name, length, dproc_root);
}



/***************
 * kdproc_read *
 *************************************************************************
 *
 * Kernel interface read bytes from a placeholder.
 *
 * @descriptor:		placeholder to read from
 * @buffer:		buffer to place (a copy of) the placeholder buffer
 * @length:		length of @buffer
 *
 * returns number of bytes actually read
 */
extern int
kdproc_read(int descriptor, char *buffer, int length)
{
  int ignore=0;

  return generic_read(buffer, &buffer, 0, length, &ignore, (void*)descriptor);
}



/***************
 * kdproc_read *
 *************************************************************************
 *
 * Kernel interface read bytes from a placeholder.
 *
 * @descriptor:		placeholder to read from
 * @buffer:		buffer to place (a copy of) the placeholder buffer
 * @length:		length of @buffer
 *
 * returns number of bytes actually read
 */
extern int
kdproc_write(int descriptor, const char *buffer, unsigned long length)
{
  return _write(descriptor, buffer, length, 1);
}


module_init(dproc_init);
module_exit(dproc_exit);

EXPORT_SYMBOL(kdproc_create_placeholder);
EXPORT_SYMBOL(kdproc_delete_placeholder);
EXPORT_SYMBOL(kdproc_create_node);
EXPORT_SYMBOL(kdproc_delete_node);
EXPORT_SYMBOL(kdproc_read);
EXPORT_SYMBOL(kdproc_write);
