/* -*-C-*-
*******************************************************************************
*
* File:         imgClient.c
* RCS:          $Id: $
* Authors:      original implementation:
*		Fabian E. Bustamante
*		extension to distributed image streaming:
*		Christian Poellabauer
*               Systems Research Group
*               College of Computing
*               Georgia Institute of Technology
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
* (C) Copyright 2001, Georgia Institute of Technology, all rights reserved.
*
*******************************************************************************
*/

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>
#include <linux/qfabric.h>
#include <linux/ecalls.h>

#include "imgHandler.h"
#include "bitio.h"
#include "errhand.h"
#include "main.h"


#define CLIENT_CROP    1
#define CLIENT_GREY    2
#define CLIENT_EDGE    3
#define CLIENT_ENLARGE 4
#define CLIENT_REDUCE  5
#define SERVER_CROP    6
#define SERVER_GREY    7
#define SERVER_EDGE    8
#define SERVER_ENLARGE 9
#define SERVER_MIX     10
#define SERVER_CONCAT  11
#define SERVER_REDUCE  12
#define HUFFMAN        13
#define LZ77           14

#define APP_ID		"PPMSTREAM"
#define GROUP_ID	"VIDEO"

#define DEBUG 1

struct down_info {
	int type;
	int pid;
};

typedef struct node_t {
	struct node_t *next;
	void *data;
} Node, *PNode;

typedef struct list_t {
	PNode head;
	PNode tail;
	int size;
} List, *PList;


int rate     = 0;
int loop     = 0;
int filterid = 0;
char *server_name = NULL;
char *local_file  = NULL;
char *ppm_file1   = NULL;
char *ppm_file2   = NULL;
FILE *filter_file = NULL;
PList filter_list;

/* for jitter average: */
long jitter1 = 0;
long jitter2 = 0;
long jitter3 = 0;
long avg_jitter = 0;
int myindex = 0;
long period;
struct timeval previous, mynow;


/*----------------------------------------------------------------------------*/

/*
 * usage: Prints out 'usage' string
 *
 * Parameters: target 'stream' and 'status' to quit with
 *
 * Returns: -
 *
 */
 
static void usage (FILE* stream, int status)
{
  fprintf(stream, "Usage: imgClient -s <server> -l <local file> -f [filter_id] -r <rate> -i <iterations> -p [ppm]\n"
	  "\t 0  -> No Filter \n"
	  "\t 1  -> Client Crop \n"
	  "\t 2  -> Client Grey \n"
	  "\t 3  -> Client Edge \n"
	  "\t 4  -> Client Enlarge \n"
	  "\t 5  -> Client Reduce \n"
	  "\t 6  -> Server Crop \n"
	  "\t 7  -> Server Grey \n"
	  "\t 8 -> Server Edge \n"
	  "\t 9 -> Server Enlarge \n"
	  "\t 10 -> Server Mix \n"
	  "\t 11 -> Server Cat \n"
	  "\t 12 -> Server Reduce \n"
	  "\t 13 -> Huffman Coding\n"
	  "\t 14 -> LZ77 Coding\n");
  exit(status);
} /* end usage */


/*------------------------------------------------------------------------------*/

static void parse_args (int argc, char **argv)
{
  int optchar;

  if (argc < 12) { usage (stderr, EXIT_FAILURE); }

  while ((optchar = getopt (argc, argv, "s:l:f:r:i:p:")) != EOF) {
  	switch (optchar) {
		case 's' :	server_name = strdup (optarg);
				#ifdef DEBUG
					fprintf (stdout, "Server: %s\n", server_name);
				#endif
				break;
		case 'l' :	local_file = strdup (optarg);
				#ifdef DEBUG	
					fprintf (stdout, "Local File: %s\n", local_file);
				#endif
				break;
		case 'f' :    	if ((filter_file = fopen(optarg, "r")) == NULL) {
				        perror ("Unable to open filter file!");
				        exit (1);
				}
				#ifdef DEBUG
					fprintf (stdout, "Filter File: %s\n", optarg);
				#endif
				/*
				switch (atoi (optarg)) {
					case 1  : filterid |= CLIENT_CROP;    break;
					case 2  : filterid |= CLIENT_GREY;    break;
					case 3  : filterid |= CLIENT_EDGE;    break;
					case 4  : filterid |= CLIENT_ENLARGE; break;
					case 5  : filterid |= CLIENT_REDUCE;  break;
					case 6  : filterid |= SERVER_CROP;    break;
					case 7  : filterid |= SERVER_GREY;    break;
					case 8  : filterid |= SERVER_EDGE;    break;
					case 9  : filterid |= SERVER_ENLARGE; break;
					case 10 : filterid |= SERVER_MIX;     break;
					case 11 : filterid |= SERVER_CONCAT;  break;
					case 12 : filterid |= SERVER_REDUCE;  break;
					case 13 : filterid |= HUFFMAN;        break;
					case 14 : filterid |= LZ77;           break;
					default : fprintf (stdout, "Error! Unknown filter!");
						  exit (EXIT_FAILURE);
						  break;
				}
				#ifdef DEBUG
					fprintf (stdout, "Filter ID: %d\n", filterid);
				#endif
				*/
				break;
		case 'r' :	rate = atoi (optarg);
				period = 1000000/rate;
				#ifdef DEBUG
					fprintf (stdout, "Rate: %d\n", rate);
				#endif
				break;
		case 'i' :	loop = atoi (optarg);
				#ifdef DEBUG
					fprintf (stdout, "Iterations: %d\n", loop);
				#endif
				break;
		case 'p' :	if (ppm_file1 == NULL) {
					ppm_file1 = strdup (optarg);
				}
				else {
					ppm_file2 = strdup (optarg);
				}
				break;
				
		case ':' :      fprintf (stdout, "Usage error! Missing a necessary parameter!\n");
				usage (stderr, EXIT_FAILURE);
				break;
				
		default  :	fprintf (stdout, "Usage error! Unknown command line parameter!\n");
				usage (stderr, EXIT_FAILURE);
				break;
	}
  }
}

/*------------------------------------------------------------------------------*/

static PList init_list (void)
{
	PList new_list;

	if ((new_list = malloc (sizeof (List))) == NULL) {
		fprintf (stderr, "List allocation error!\n");
		exit(1);
	}
	new_list->head = new_list->tail = NULL;
	new_list->size = 0;

	return new_list;
}


static void add_to_list (PList list, void *data, int size)
{
	PNode new_node;

	if ((new_node = malloc (sizeof (Node))) == NULL) {
		fprintf (stderr, "Filter node allocation error!\n");
		exit (1);
	}

	new_node->next = NULL;
	if ((new_node->data = malloc (sizeof (char) * size)) == NULL) {
		fprintf (stderr, "Filter data allocation error!\n");
		exit(1);
	}
	memcpy (new_node->data, data, size);

	if (list->head == NULL) {
		list->head = list->tail = new_node;
	}
	else {
		list->tail->next = new_node;
		list->tail = new_node;
	}
	list->size++;
}


static void *get_list_node (PList list, int position)
{
	int i = 0;
	PNode helper = list->head;
	
	while (i < list->size && i != position) {
		helper = helper->next;
		i++;
	}
	if (helper != NULL) {
		return helper->data;
	}
	else {
		return NULL;
	}
}



static void free_list_helper (PNode node)
{
	if (node->next != NULL) {
		free_list_helper (node->next);
		free (node);
	}
}



static void free_list (PList list)
{
	free_list_helper (list->head);
	free (list);
}




/*------------------------------------------------------------------------------*/

static void end_transmission (int socket)
{
	struct stream_response filter_length;

	filter_length.len = -999;
	filter_length.len = htonl (filter_length.len);

	send (socket, &filter_length, sizeof (struct stream_response), 0);
	
}



static void transmit_filters (int socket)
{
	int type;
	char type_string[5];
	char *file_entry;


	#ifdef DEBUG
		fprintf (stdout, "About to transmit filters...\n");
		fflush (stdout);
	#endif
	
	if (fscanf (filter_file, "%s", type_string) == EOF) {
		fclose (filter_file);
		end_transmission (socket);
		return;
	}
	#ifdef DEBUG
		fprintf (stdout, "Type String: %s\n", type_string);
		fflush (stdout);
	#endif
	if (strcmp (type_string, "AND") == 0) {
		type = AND_TYPE;
	}
	else {
		type = OR_TYPE;
	}
	
	#ifdef DEBUG
		fprintf (stdout, "AND/OR: %d\n", type);
	#endif

	filter_list = init_list();

		
	while (fscanf (filter_file, "%s\n", file_entry) != EOF) {
		struct stream_response filter_length;
		struct filter_packet_t filter_data;

		#ifdef DEBUG
			fprintf (stdout, "File Entry: %s\n", file_entry);
			fprintf (stdout, "Entry Length: %d\n", strlen (file_entry));
		#endif

		filter_length.len = htonl (strlen (file_entry));
	
		filter_data.type = htonl (type);
		memcpy (filter_data.data, file_entry, strlen (file_entry));
		filter_data.data[strlen(file_entry)] = '\0';
		add_to_list (filter_list, filter_data.data, strlen (file_entry));

		#ifdef DEBUG
			fprintf (stdout, "Sending Data: %s\n", (char *) filter_data.data);
		#endif

		send (socket, &filter_length, sizeof (struct stream_response), 0);
	        send (socket, &filter_data, sizeof (struct filter_packet_t), 0);
	}

	fclose (filter_file);
	end_transmission (socket);
	return;
}



/*------------------------------------------------------------------------------*/

int main (int argc, char* argv[])
{ 
  int i;
  int count;
  int bytes_recv;
  int bytes_sent;
  int compressed = 0;
  int total_image_bytes;
  struct tm *now;
  time_t nowt;
  char header[20];
  Raw_data inImage1, outImage, tmpImage;
  int inCol = 0;
  int inRow = 0;
  int outWidth = 100; 
  int outHeight = 100;
  int sfd;
  int enlargeFactor = 1;
  int reduceFactor = 1;
  int maxppmval;
  int fd;
  struct sockaddr_in server;
  struct hostent *hp;
  struct stream_request my_request;
  int eof_rcvd = 0;
  char *newimage;
  struct stream_response ack;
  pid_t my_pid;
  char filename1[20];
  char filename2[20];
  BIT_FILE *fd_in;
  int fd_in2;
  FILE *fd_out;
  struct ecall_down downcall;
  struct ecall_up upcall;
  int spid;
  struct down_info down;
  
  parse_args (argc, argv);

  ECalls_start(0);
  
  /* Register for QoS managememt */
  qos_configure(APP_ID, GROUP_ID);

  downcall.size1 = 0; 
  downcall.size2 = 0; 
  if (ECalls_subscribe("QoS", &downcall, NULL, &upcall, NULL) < 0) {
	printf("ECalls Error!\n");
	return -1;
  }

  previous.tv_sec = -1;
  previous.tv_usec = -1;
 
  /* Create connection to server */
  fd = socket(PF_INET, SOCK_STREAM, 0); 
  if (fd == -1) {
    perror("Failed opening socket.\n");    
    exit(1);
  }
#ifdef DEBUG
  printf("Socket connection created\n");
#endif
  bzero((char *) &server, sizeof(server));
  server.sin_family = PF_INET;
  if ((hp = gethostbyname(server_name)) == NULL) {
    perror("Unknown server.\n");    
    exit(1);
  }
  bcopy(hp->h_addr, &server.sin_addr, hp->h_length);
  server.sin_port = htons ((u_short) SERVER_PORT);
  if (connect(fd, (const struct sockaddr *) &server, sizeof(server)) == -1) {
    perror("Failed connecting to server.\n");    
    exit(1);
  }
#ifdef DEBUG
  printf("Socket connect successful\n");
#endif

  time(&nowt);
  now = localtime(&nowt);

  transmit_filters (fd);

  /* Submit service request */
  my_request.filter_id = htonl (filterid);
  my_request.rate = htonl (rate);
#ifdef DEBUG
  printf("Rate: %d\n", ntohl (my_request.rate));
  printf("name: %s\n", ppm_file1);
  printf("Filter #: %d\n", filterid);
#endif
  strcpy((char *) &my_request.FileName1, ppm_file1);
  my_request.FileNameLength1 = strlen (ppm_file1);
  my_request.FileName1[my_request.FileNameLength1] = '\0';
  if (ppm_file2 != NULL) {
    strcpy((char *) &my_request.FileName2, ppm_file2);
    my_request.FileNameLength2 = strlen (ppm_file2);
    my_request.FileName2[my_request.FileNameLength2] = '\0';
  } else {
    my_request.FileNameLength2 = 0;
  }
  my_request.loop = htonl (loop);
  if (my_request.loop <= 0) my_request.loop = htonl(1);
  
#ifdef DEBUG 
  printf("Send request to server\n");
#endif 
  write(fd, (const void *)&my_request, sizeof (struct stream_request));

  bytes_recv = recv (fd, &spid, sizeof(int),0);
  down.type = 1;
  down.pid = spid;
  ECalls_submit_event(&downcall, &down, sizeof(down), SYSCALL);

  /* Read images from socket */
  count = 0;
  while (eof_rcvd == 0) {
	  struct stream_response iheader;

    if ((bytes_recv = recv (fd, &iheader, sizeof(struct stream_response),0)) <= 0)
	continue;
//#ifdef DEBUG 
//  printf("Received response: %d in %d bytes\n", (int) ntohl (iheader.len), bytes_recv);
//#endif 
    if (((int) ntohl (iheader.len)) <=0) {
      if (((int) ntohl (iheader.len)) == -999) {
	fprintf (stdout, "EOF received!!!\n");
	fflush (stdout);
	eof_rcvd = 1;
      }
      continue;
    }

    /* Open new file */
#ifdef DEBUG 
    printf("Open file %s\n", local_file);
#endif 
    unlink(local_file);
    if ((sfd = open(local_file, O_CREAT|O_WRONLY, S_IRWXU|S_IRGRP|S_IROTH)) < 0) {
	perror ("Unable to open local file!");
	exit (1);
    }

    newimage = malloc(ntohl (iheader.len));
    total_image_bytes = 0;
    while (total_image_bytes != ntohl (iheader.len)) {
    	if ((bytes_recv = recv(fd, newimage, ntohl (iheader.len),0)) <=0 )
		continue;
	total_image_bytes += bytes_recv;
    }
    fprintf (stdout, "Received an image: %d bytes!\n", total_image_bytes);
    fflush (stdout);
    if (!(filterid & HUFFMAN) && !(filterid & LZ77)) {
	#ifdef DEBUG
	    fprintf (stdout, "Getting 'in' image sizes...\n");
	#endif
      sscanf(newimage, "%c%c\n%d %d\n%d\n",
	&(inImage1.ppm1), &(inImage1.ppm2),
	&(inImage1.width), &(inImage1.height),
	&maxppmval);
      inImage1.buff = (u_char *) (newimage + 20);
      inImage1.size = ntohl (iheader.len) - sizeof(header);
	#ifdef DEBUG
	    fprintf (stdout, "New Image Size: %d\n", inImage1.size);
        #endif
    }

	// Filter image
    outImage.buff = NULL;
    for (i = 0; i < filter_list->size; i++) {
	    int filter_type;
	    char filter_params[32];
	    char *filter_info;

	    filter_info = (char *) get_list_node (filter_list, i);
	    sscanf (filter_info, "%d:%s", &filter_type, filter_params);
	    #ifdef DEBUG
	    	fprintf (stdout, "Filter Type: %d\nList Size: %d\n", filter_type, filter_list->size);
	    	fflush (stdout);
	    #endif

	    switch (filter_type) {
		    case SERVER_CROP    : break;
		    case SERVER_GREY    : break;
		    case SERVER_EDGE    : break;
		    case SERVER_ENLARGE : break;
		    case SERVER_REDUCE  : break;
		    case SERVER_MIX	: break; 
		    case SERVER_CONCAT	: break;
					  
		    case CLIENT_CROP    : sscanf (filter_params, "%d:%d:%d:%d", 
							  &inCol, &inRow, &outWidth, &outHeight);
					  if (inCol < 0 || inCol > inImage1.width) inCol = 0; 
        				  if (inRow < 0 || inRow > inImage1.height) inRow = 0; 
					  if (outWidth < 0 || outWidth > (inImage1.width - inCol)) 
            					outWidth = inImage1.width - inCol; 
        				  if (outHeight < 0 || outHeight > (inImage1.height - inRow)) 
					        outHeight = inImage1.height - inRow; 
					  outImage.size = IMGSIZECOLOR(outWidth, outHeight);
					  if (outImage.buff == NULL) {
						outImage.buff = (u_char*) malloc(outImage.size);
					  }
					  else {
						outImage.buff = realloc (outImage.buff, outImage.size);
					  }
					  if (outImage.buff == NULL) {
						perror ("Memory allocation error.");
						exit(1);
        				  }
        				  if (cropImage (&inImage1, &outImage, inRow, inCol, 
								        outWidth, outHeight) < 0) {
						perror ("Error filtering image.");
						exit(1);
        				  }
					  break;
					  
		    case CLIENT_GREY    : outImage.size = IMGSIZEGREY(inImage1.width, inImage1.height);
        				  if (outImage.buff == NULL) {
						outImage.buff = (u_char*) malloc(outImage.size);
					  }
					  else {
						outImage.buff = realloc (outImage.buff, outImage.size);
					  }
					  if (outImage.buff == NULL) {
						fprintf (stderr, "Requested %d bytes...\n", outImage.size);
						perror ("Memory allocation error.");
						exit(1);
					  }
    			                  if (greyImage(&inImage1, &outImage) < 0) {
					  	perror ("Error filtering image.");
					        exit(1);
        				  }
					  break;
					  
		    case CLIENT_EDGE    : tmpImage.size = IMGSIZEGREY(inImage1.width, inImage1.height);
					  tmpImage.buff = (u_char*) malloc(tmpImage.size);
				          if (tmpImage.buff == NULL) {
					        perror ("Memory allocation error.");
					        exit(1);
				          }
    				          if (greyImage(&inImage1, &tmpImage) < 0) {
					        perror ("Error filtering image.");
					        exit(1);
					  }

				          outImage.size = IMGSIZEGREY(tmpImage.width, tmpImage.height);
					  if (outImage.buff == NULL) {
				          	outImage.buff = (u_char*) malloc(outImage.size);
					  }
					  else {
						outImage.buff = realloc (outImage.buff, outImage.size);
					  }
				          if (outImage.buff == NULL) {
					  	perror ("Memory allocation error.");
					        exit(1);
					  }
    				          if (edgeImage(&tmpImage, &outImage) < 0) {
					          perror ("Error filtering image.");
            					  exit(1);
					  }
					  break;
					  
		    case CLIENT_ENLARGE : sscanf (filter_params, "%d", &enlargeFactor);
					  if (enlargeFactor < 0) enlargeFactor = 1;
        				  outImage.size = IMGSIZECOLOR((inImage1.width * enlargeFactor), 
	  			      				       (inImage1.height * enlargeFactor));
					  if (outImage.buff == NULL) {
					  	outImage.buff = (u_char*) malloc(outImage.size);
					  }
					  else {
						outImage.buff = realloc (outImage.buff, outImage.size);
					  }
      
					  if (outImage.buff == NULL) {
						perror ("Memory allocation error.");
						exit(1);
					  }
					  if (enlargeImage(&inImage1, &outImage, enlargeFactor) < 0) {
						perror ("Error filtering image.");
						exit(1);
					  }
					  break;
					  
		    case CLIENT_REDUCE  : sscanf (filter_params, "%d", &reduceFactor);
					  if (reduceFactor < 0) reduceFactor = 1;
					  outImage.size = IMGSIZECOLOR((inImage1.width / reduceFactor), 
				      					(inImage1.height / reduceFactor));
				          outImage.buff = (u_char*) malloc(outImage.size);
      					  if (outImage.buff == NULL) {
					        perror ("Memory allocation error.");
					        exit(1);
					  }
    
				          if (reduceImage(&inImage1, &outImage, reduceFactor) < 0) {
					        perror ("Error filtering image.");
					        exit(1);
					  }
					  break;  
					  
		    case HUFFMAN        : my_pid = getpid();
					  sprintf(filename1, "%s-%d-1", "temp", my_pid);
				          sprintf(filename2, "%s-%d-2", "temp", my_pid);
				          fd_in2 = open(filename1, O_CREAT|O_WRONLY, S_IRWXU|S_IRGRP|S_IROTH);
			                  write(fd_in2, newimage, iheader.len);
					  close(fd_in2);
			                  fd_in = OpenInputBitFile(filename1);
					  fd_out = fopen(filename2, "wb");
					  ExpandFile(fd_in, fd_out, 0, NULL);
					  CloseInputBitFile(fd_in);
					  fclose(fd_out);
					  close(sfd);
				          rename(filename2, local_file);
					  unlink(filename1); 
					  compressed = 1;
					  break;
					  
		    case LZ77           : my_pid = getpid();
					  sprintf(filename1, "%s-%d-1", "temp", my_pid);
					  sprintf(filename2, "%s-%d-2", "temp", my_pid);
					  fd_in2 = open(filename1, O_CREAT|O_WRONLY, S_IRWXU|S_IRGRP|S_IROTH);
					  write(fd_in2, newimage, iheader.len);
					  close(fd_in2);
					  fd_in = OpenInputBitFile(filename1);
					  fd_out = fopen(filename2, "wb");
				          ExpandFile2(fd_in, fd_out, 0, NULL);
					  CloseInputBitFile(fd_in);
					  fclose(fd_out);
					  close(sfd);
					  rename(filename2, local_file);
					  unlink(filename1);
					  compressed = 1;
					  break;

		    default             : fprintf (stderr, "Unknown filter code!\n");
					  exit (1);
					  break;
	    }

	    if ((i < (filter_list->size - 1)) && (filter_type < SERVER_CROP)) {
		    free (inImage1.buff);
		    inImage1.size = outImage.size;
		    if ((inImage1.buff = malloc (outImage.size)) == NULL) {
			    fprintf (stderr, "Unable to reallocate images!\n");
			    exit (1);
		    }
		    memcpy (inImage1.buff, outImage.buff, outImage.size);
		    free (outImage.buff);
		    outImage.buff = NULL;
	    }
    }

    if (!compressed) {
	    if (outImage.buff != NULL) {
		sprintf(header, "%c%c\n%d %d\n%d\n", outImage.ppm1, outImage.ppm2, 
		    	outImage.width, outImage.height, outImage.maxval);
#ifdef DEBUG
	    fprintf (stdout, "Writing image to disk...%d bytes\n", outImage.size);
	    fflush (stdout);
#endif
		write(sfd, header, sizeof(header));
		write(sfd, outImage.buff, outImage.size);
	    }
	    else {
		sprintf(header, "%c%c\n%d %d\n%d\n", inImage1.ppm1, inImage1.ppm2, 
		 	inImage1.width, inImage1.height, inImage1.maxval);
#ifdef DEBUG
	    fprintf (stdout, "Writing image to disk...%d bytes\n", inImage1.size);
	    fflush (stdout);
#endif
	    	write(sfd, header, sizeof(header));
	        write(sfd, inImage1.buff, inImage1.size);
	    }
	    free (outImage.buff);
      	    close(sfd);
    }

    ack.len = 12345 + count;
    ack.len = htonl (ack.len);
    bytes_sent = send (fd, &ack, sizeof (struct stream_response), 0);
    #ifdef DEBUG
    	fprintf (stdout, "Sending server acknowledgement...%d bytes\n", bytes_sent);
    #endif
    count++;
    
    free(newimage);

    /* jitter, averaged over last 3 frames. */
    gettimeofday(&mynow, NULL);
    if (((previous.tv_usec == -1) && (previous.tv_sec == -1)) || (myindex == 0))
    {
	previous.tv_sec = mynow.tv_sec;
	previous.tv_usec = mynow.tv_usec;
	myindex = 1;
    }
    else {
      if (myindex == 1) {
        jitter1 = (mynow.tv_sec - previous.tv_sec)*1000000;
	if (jitter1 == 0) 
	  jitter1 = mynow.tv_usec - previous.tv_usec;
  	else 
	  jitter1 = jitter1 + mynow.tv_usec + (1000000-previous.tv_usec) 
		-1000000;
	jitter1 = jitter1 - period;
        myindex++;
      } else if (myindex == 2) {
        jitter2 = (mynow.tv_sec - previous.tv_sec)*1000000;
	if (jitter2 == 0) 
	  jitter2 = mynow.tv_usec - previous.tv_usec;
  	else 
	  jitter2 = jitter2 + mynow.tv_usec + (1000000-previous.tv_usec) 
		-1000000;
	jitter2 = jitter2 - period;
        myindex++;
      } else if (myindex == 3) {
        jitter3 = (mynow.tv_sec - previous.tv_sec)*1000000;
	if (jitter3 == 0) 
	  jitter3 = mynow.tv_usec - previous.tv_usec;
  	else 
	  jitter3 = jitter3 + mynow.tv_usec + (1000000-previous.tv_usec)
		-1000000;
	jitter3 = jitter3 - period;
	myindex = 1;
      }
      avg_jitter = jitter1 + jitter2 + jitter3;
      avg_jitter = avg_jitter / 3; /* if + too slow, - too fast */
      ECalls_submit_event(&downcall, &avg_jitter, sizeof(long), SYSCALL);
      previous.tv_sec = mynow.tv_sec;
      previous.tv_usec = mynow.tv_usec;
    }

    #ifdef DEBUG
	fprintf (stdout, "End of loop iteration...\n");
	fflush (stdout);
    #endif

  }

  free_list (filter_list);

  close (fd);

  ECalls_cancel(&downcall, &upcall);  
  return 0;
} /* end main */
