
#ifdef MODULE
#ifndef __KERNEL__
#define __KERNEL__
#endif
#endif

#ifndef MODULE
#include "config.h"
#include "ltdl.h"
#include <stdio.h>
#include <string.h>
#include <assert.h>
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#include <io.h>
#include <atl.h>
#include "cm.h"
#include "gen_thread.h"
#include "cm_internal.h"
#else
#include <linux/wait.h>
#include "config.h"
#include "kcm.h"
#include "library.h"
#include "../include/io.h"
#include "../include/atl.h"
#include "cm.h"
#include "cm_internal.h"
#include "assert.h"

#define printf         printk
#define perror         printk
#define fprintf(fmt, args...) printk (args) 
#define malloc         (void *)DAllocMM
#define free(a)        DFreeMM((addrs_t)a)
#define realloc(a,b)   (void *)DReallocMM((addrs_t)a,b)
#define exit           sys_exit
#define qsort          _quicksort
#define getpid()       current->pid
#define signal(a,b)    
#define gethostname    sys_gethostname
#define gethostbyname  lookup_name
#define select         sys_select
#define thr_condition_signal(a)
#define thr_condition_free(a)

static DECLARE_WAIT_QUEUE_HEAD(cm_queue);
#endif

typedef struct _CMCondition {
    CMCondition next;
    int condition_num;
    int waiting;
    int signaled;
    int failed;
    thr_condition_t cond_condition;
    CMConnection conn;
    void *client_data;
} CMCondition_s;

static int cm_control_debug_flag = -1;

static void set_debug_flag(cm)
CManager cm;
{
    if (cm_control_debug_flag == -1) {
	if (CMtrace_on(cm, CMLowLevelVerbose)) {
	    cm_control_debug_flag = 1;
	} else {
	    cm_control_debug_flag = 0;
	}
    }
}

#ifdef MODULE
void cm_wake_up(unsigned long data)
{
  wake_up_interruptible(&cm_queue);
}
#endif

static CMCondition
CMCondition_find(cl, condition)
CMControlList cl;
int condition;
{
    CMCondition next = cl->condition_list;
    while (next != NULL) {
	if (next->condition_num == condition) {
	    return next;
	}
	next = next->next;
    }
    fprintf(stderr, "Serious internal error.  Use of condition %d, no longer in control list\n", condition);
    return NULL;
}


extern int
CMCondition_get(cm, conn)
CManager cm;
CMConnection conn;
{
    CMControlList cl = cm->control_list;
    CMCondition cond = CMmalloc(sizeof(CMCondition_s));
    set_debug_flag(cm);
    CMControlList_lock(cl);
    cond->next = cl->condition_list;
    cl->condition_list = cond;
    cond->condition_num = cl->next_condition_num++;
    cond->conn = conn;
    if (cl->next_condition_num >= 0xffffff) {
	/* recycle at  (16 M - 1) [ Caution on number reuse ] */
	cl->next_condition_num = 0;
    }
    cond->waiting = 0;
    cond->signaled = 0;
    cond->failed = 0;
    cond->cond_condition = NULL;
#ifndef MODULE
    if (cl->has_thread) {
	cond->cond_condition = thr_condition_alloc();
    }
#endif
    CMControlList_unlock(cl);
    return cond->condition_num;
}

static void
CMCondition_trigger(cond, cl) 
CMCondition cond;
CMControlList cl;
{
    if (cm_control_debug_flag) {
	printf("CMLowLevel Triggering CMcondition %d\n", cond->condition_num);
    }
    if (cond->waiting) {
	if (cl->has_thread) {
	    if (cm_control_debug_flag) {
		printf("CMLowLevel Triggering CMcondition %d, thr_cond %lx\n", 
		       cond->condition_num, (long)cond->cond_condition);
	    }
	    thr_condition_signal(cond->cond_condition);
	}
    }
    if (cm_control_debug_flag) {
	printf("CMLowLevel After trigger for CMcondition %d\n", 
	       cond->condition_num);
    }
}

extern void
CMconn_fail_conditions(conn)
CMConnection conn;
{
    CMControlList cl = conn->cm->control_list;
    CMCondition cond_list;
    set_debug_flag(conn->cm);
    CMControlList_lock(cl);
    cond_list = cl->condition_list;
    while(cond_list != NULL) {
	if (cond_list->conn == conn) {
	    cond_list->failed = 1;
	    CMCondition_trigger(cond_list, cl);
	}
	cond_list = cond_list->next;
    }
    CMControlList_unlock(cl);
}

void
CMCondition_destroy(cl, condition)
CMControlList cl;
int condition;
{
    CMCondition cond = NULL, prev = NULL;
    CMCondition next = NULL;

    if (cl->condition_list->condition_num == condition) {
	cond = cl->condition_list;
	cl->condition_list = cl->condition_list->next;
    } else {
	prev = cl->condition_list;
	next = cl->condition_list->next;
	while (next != NULL) {
	    if (next->condition_num == condition) {
		cond = next;
		prev->next = next->next;
		break;
	    }
	    prev = next;
	    next = next->next;
	}
    }
    if (cond == NULL) {
	fprintf(stderr, "Serious internal error.  Use of condition %d, no longer in control list\n", condition);
    } else {
	/* free internal elements */
	if (cond->cond_condition) {
	    thr_condition_free(cond->cond_condition);
	}
	CMfree(cond);
    }
}

extern int
CMCondition_has_signaled(cm, condition)
CManager cm;
int condition;
{
    int retval;
    CMCondition cond;
    CMControlList cl = cm->control_list;
    set_debug_flag(cm);

    CMControlList_lock(cl);
    
    cond = CMCondition_find(cl, condition);
    retval = cond->signaled;
    
    CMControlList_unlock(cl);
    return retval;
}

extern int
CMCondition_has_failed(cm, condition)
CManager cm;
int condition;
{
    int retval;
    CMCondition cond;
    CMControlList cl = cm->control_list;
    set_debug_flag(cm);

    CMControlList_lock(cl);
    
    cond = CMCondition_find(cl, condition);
    retval = cond->failed;
    
    CMControlList_unlock(cl);
    return retval;
}

extern int
CMCondition_wait(cm, condition)
CManager cm;
int condition;
{
    CMCondition cond;
    CMControlList cl = cm->control_list;
    int result;
#ifdef MODULE
    int counter = 0;
    int expire = 0;
//    DECLARE_WAITQUEUE(cm_wait, current);
#endif

    set_debug_flag(cm);
    if (cm_control_debug_flag) {
	printf("CMLowLevel Waiting for CMcondition %d\n", condition);
    }
    CMControlList_lock(cl);
    if (cm_control_debug_flag) {
	printf("CMLowLevel locked cl\n");
    }
    cond = CMCondition_find(cl, condition);

    if (cond->signaled) {
	if (cm_control_debug_flag) {
	    printf("CMcondition %d already signalled\n", condition);
	}
	CMControlList_unlock(cl);
	return 1;
    }
    if (cond->failed) {
	if (cm_control_debug_flag) {
	    printf("CMcondition %d already failed\n", condition);
	}
	CMControlList_unlock(cl);
	return 0;
    }
    cond->waiting++;
    if (cm_control_debug_flag) {
	printf("CMLowLevel In condition wait, server thread = %lx\n", 
	       (long)cl->server_thread);
    }
#ifndef MODULE
    if (!cl->has_thread) {
	while (!(cond->signaled || cond->failed)) {
	    if (cm_control_debug_flag) {
		printf("CMLowLevel  Polling for CMcondition %d\n", condition);
	    }
	    CMControlList_unlock(cl);
	    CMcontrol_list_wait(cl);
	    CMControlList_lock(cl);
	}
	if (cm_control_debug_flag) {
	    printf("CMLowLevel  after Polling for CMcondition %d\n", condition);
	}
	/* the poll and handle will set cl->server_thread, restore it */
	cl->server_thread = NULL;
	if (cm_control_debug_flag) {
	    printf("CMLowLevel  In condition wait, reset server thread = %lx\n", 
		   (long)cl->server_thread);
	}
    } else if (thr_thread_self() == cl->server_thread) {
#else
    if (1) {
#endif
	/* we're the server thread */
#ifdef MODULE
        CMControlList_unlock(cl);
	CMcontrol_list_wait(cl);
	CMControlList_lock(cl);
#endif
	while (!(cond->signaled || cond->failed)) {
	    if (cm_control_debug_flag) {
		printf("CMLowLevel polling for CMcondition %d\n", condition);
	    }
	    CMControlList_unlock(cl);
	    CMcontrol_list_wait(cl);
#ifdef MODULE
/*
	    del_timer(&cm->cm_timer);
	    cm->cm_timer.function = cm_wake_up;
	    cm->cm_timer.data = 0;
*/
	    if (expire <= jiffies)
	      expire = jiffies + PERIOD;
/*
   	    cm->cm_timer.expires = expire;
	    add_timer(&cm->cm_timer);
            current->qfabric->sleep_queue = (unsigned long)&cm_queue;
            current->state = TASK_INTERRUPTIBLE;
	    add_wait_queue(&cm_queue,&cm_wait);
            schedule();
	    current->state = TASK_RUNNING;
            current->qfabric->sleep_queue = (unsigned long)0;
 	    remove_wait_queue(&cm_queue,&cm_wait);
	    del_timer(&cm->cm_timer);
*/
	    CMtask_sleep(PERIOD);
	    if (jiffies >= expire)
	      counter++;
 	    if ((counter > 20) || (current->qfabric->sw_pending == -1)) {
		printk("Wait failed!\n");
		return 0;
	    }
#endif
	    CMControlList_lock(cl);
	}
#ifdef MODULE
    }
#else
    } else {
	/* some other thread is the server thread */
	if (!gen_thr_initialized()) {
	    fprintf(stderr, "Gen_Thread library not initialized.\n");
	    return 0;
	}
	if (cm_control_debug_flag) {
	    printf("CMLowLevel Waiting for CMcondition %d, thr_cond %lx\n", 
		   condition, (long)cond->cond_condition);
	}
	thr_condition_wait(cond->cond_condition, cl->list_mutex);
	if (cm_control_debug_flag) {
	    printf("CMLowLevel After wait for CMcondition %d, thr_cond %lx\n", 
		   condition, (long)cond->cond_condition);
	}
    }
#endif
    result = cond->signaled;
    CMCondition_destroy(cl, condition);
    CMControlList_unlock(cl);
    if (cm_control_debug_flag) {
	printf("CMLowLevel Return from wait CMcondition %d\n", condition);
    }
    return result;
}

extern void
CMCondition_signal(cm, condition)
CManager cm;
int condition;
{
    CMCondition cond;
    CMControlList cl = cm->control_list;
    set_debug_flag(cm);
    CMControlList_lock(cl);
    cond = CMCondition_find(cl, condition);
    cond->signaled = 1;
    CMCondition_trigger(cond, cl);
    CMControlList_unlock(cl);
}

extern void
CMCondition_set_client_data(cm, condition, client_data)
CManager cm;
int condition;
void *client_data;
{
    CMCondition cond;
    CMControlList cl = cm->control_list;
    set_debug_flag(cm);
    CMControlList_lock(cl);
    cond = CMCondition_find(cl, condition);
    cond->client_data = client_data;
    CMControlList_unlock(cl);
}

extern void *
CMCondition_get_client_data(cm, condition)
CManager cm;
int condition;
{
    CMCondition cond;
    void *client_data;
    CMControlList cl = cm->control_list;
    set_debug_flag(cm);
    CMControlList_lock(cl);
    cond = CMCondition_find(cl, condition);
    client_data = cond->client_data;
    CMControlList_unlock(cl);
    return client_data;
}
