
/*
 *   kcm.c
 *
 *   Copyright (C) 2000 Christian Poellabauer
 *
 */

/* 
 *   Kernel-loadable module for event channel support
 */

/* Required definitions. */
#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif

/* Various required header files. */
#include <linux/module.h> 
#include <linux/kernel.h> 
#include <linux/tty.h>
#include <linux/unistd.h>
#include <linux/socket.h>
#include <linux/poll.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <asm/current.h>
#include <asm/uaccess.h>
//#include <linux/aio.h>
#include "kcm.h"
#include "../include/io.h"
#include "cm.h"
#include "cm_internal.h"

MODULE_AUTHOR("CERCS - GATECH");
MODULE_DESCRIPTION("CM Module (KECho)");
MODULE_LICENSE("GPL");

#define MAX_THREADS 1000
#define TRIES 5

/* Prototypes. */
int init_module(void);		       
void cleanup_module(void);
int (*original_syscall_cm)();
int (*original_sw_callback)();

/* arch/i386/kernel/entry.S */
extern void *sys_call_table[];
extern void *fast_sys_call_table[];

DECLARE_MUTEX(cm_sem);

int kecho_identifier;
wait_queue_head_t kecho_queue;
int timerrun_id = -1;
int destroy_timerrun = 0;
kthread_t *resubmit_threads[MAX_THREADS];
int terminate_threads = 0;

extern void cm_task_wake_up(struct task_struct *, int, int, void *, int);

int int_sw_callback(fd, pid, ptr)
int fd;
int pid;
void *ptr;
{
  struct task_struct *tsk;
 
  if (pid <= 0)
    return 0;

  tsk = find_task_by_pid(pid);

  if (tsk != NULL) {
    tsk->qfabric->sw_pending = 1;
    cm_task_wake_up(tsk, fd, pid, ptr, kecho_identifier);
  }

  return 1;
}

int int_syscall_cm(data)
void *data;
{
  syscall_attrs_cm *attrs;
  int ret;
  CManager cm = NULL;

  attrs = (syscall_attrs_cm *)data;

  if (current->qfabric == NULL)
    initialize_qfabric();

  if (strcmp("CMpoll_network", attrs->name) == 0) {
    return CMpoll_network(attrs->cm);
  }
  if (strcmp("CMretrieve_event", attrs->name) == 0) {
    ret = CMretrieve_event(attrs->cm);
    if (ret > 0) return ret;
    cm = attrs->cm;
    if ((cm == NULL) || (cm->func == NULL) || (cm->event == NULL) ||
 	(cm->event_len <= 0)) return 1;
    attrs->func = (unsigned long) cm->func;
    attrs->event_len = cm->event_len;
    copy_to_user(attrs->event, cm->event, attrs->event_len);
    attrs->client_data = cm->client_data;
    attrs->event_attrs = cm->event_attrs;
    return ret;
  }
  if (strcmp("CManager_create", attrs->name) == 0) {
    attrs->cm = CManager_create();
    return 0;
  }
  else if (strcmp("CMrun_network", attrs->name) == 0) {
    CMrun_network(attrs->cm);
    return 0;
  }
  else if (strcmp("CMsleep", attrs->name) == 0) {
    CMsleep(attrs->cm, attrs->sec);
    return 0;
  }
  else
    return -1;
}

int     (*sys_socket)      (int family, int type, int protocol);
int     (*sys_connect)     (int fd, struct sockaddr *uservaddr, int addrlen);
int     (*sys_accept)      (int fd, struct sockaddr *upeer_sockaddr, 
                            int *upeer_addrlen);
int     (*sys_bind)        (int fd, struct sockaddr *umyaddr, int addrlen);
int     (*sys_listen)      (int fd, int backlog);
int     (*sys_setsockopt)  (int fd, int level, int optname, char *optval,
                            int optlen);
int     (*sys_getsockname) (int fd, struct sockaddr *usockaddr,
                            int *usockaddr_len);
long 	(*sys_kioctl)	   (unsigned int fd, unsigned int cmd, unsigned long arg);

struct socket *my_sock_accept(struct socket *sock, int flags)
{
  struct socket *newsock;

  if (!(newsock = sock_alloc()))
    return NULL;

  newsock->type = sock->type;
  newsock->ops = sock->ops;

  if (sock->ops->accept(sock, newsock, flags) < 0) {
    sock_release(newsock);
    return NULL;
  }
  
  return newsock;
}

/* gethostname function. */
int sys_gethostname(char *name, int len)
{
  int i, errno;

  if (len < 0)
    return -EINVAL;
  down_read(&uts_sem);
  i = 1 + strlen(system_utsname.nodename);
  if (i > len)
    i = len;
  errno = 0;
  memcpy(name, system_utsname.nodename, i);
  up_read(&uts_sem);
  return errno;
}

/* getdomainname function. */
int sys_getdomainname(char *name, int len)
{
  int nlen;
  int err = -EFAULT;

  down_read(&uts_sem);

  nlen =  strlen(system_utsname.domainname) + 1;

  if (nlen < len)
    len = nlen;
  if (len > __NEW_UTS_LEN)
    goto done;
  memcpy(name, system_utsname.domainname, len);
  err = 0;
done:
  up_read(&uts_sem);
  return err;
}

void my_FD_CLR(int fd, char *fdset)
{
  fdset[fd] = 0;
}

void my_FD_SET(int fd, char *fdset)
{
  fdset[fd] = 1;
}

int my_FD_ISSET(int fd, char *fdset)
{
  if (fdset[fd] > 0) return 1;
  return 0;
}

/* Initialize module. */
int init_module(void) 
{
  int i;

  original_syscall_cm = sys_call_table[__NR_KECho_syscall_cm];
  original_sw_callback = sys_call_table[__NR_KECho_sw_callback];

  sys_call_table[__NR_KECho_syscall_cm]  = int_syscall_cm;
  sys_call_table[__NR_KECho_sw_callback]  = int_sw_callback;

  sys_kioctl	= sys_call_table[__NR_ioctl];

  init_waitqueue_head(&kecho_queue);

  for (i = 0; i < MAX_THREADS; i++) {
    resubmit_threads[i] = NULL;
  }

  return 0;
}

/* Remove module. */
void cleanup_module(void) 
{
  struct task_struct *tsk;
  int i;

  sys_call_table[__NR_KECho_syscall_cm]  = original_syscall_cm;
  sys_call_table[__NR_KECho_sw_callback]  = original_sw_callback;

  if (timerrun_id != -1) {
    destroy_timerrun = 1;
    tsk = find_task_by_pid(timerrun_id);
    if (tsk != NULL) {
      wake_up_interruptible((wait_queue_head_t *)&tsk->qfabric->kecho_queue);
      sleep_on_timeout(&kecho_queue, HZ/10);
    }
  }

  terminate_threads = 1;
  for (i = 0; i < MAX_THREADS; i++) {
  	if (resubmit_threads[i] != NULL) {
		stop_kthread(resubmit_threads[i]);
	}
  }
}

/* Socket sendto function. */
int my_sock_sendto(struct socket *sock, char *buf, int len, 
	unsigned flag, struct sockaddr *addr, int addr_len, int deadline, int flags) 
{
  struct msghdr msg;
  struct iovec iov;
  mm_segment_t oldfs;
  char address[128];
  int tries = TRIES;

  if (len == 0)
    return 0;

  if (sock == NULL) 
    return 0;

  if (sock->sk == NULL) {
    printk("sk is null!\n");
    return 0;
  }

  msg.msg_flags = MSG_DONTWAIT;
  msg.msg_name = NULL;
  msg.msg_namelen = addr_len;
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;
  msg.msg_control = NULL;
  msg.msg_controllen = 0;
  iov.iov_base = (char *)buf;
  iov.iov_len = (__kernel_size_t)len;
  if (sock->type == SOCK_SEQPACKET)
    msg.msg_flags |= MSG_EOR;
  if (addr) {
    memcpy((void *)address, (void *)addr, addr_len);
    msg.msg_name = address;
  }

  oldfs = get_fs();
  set_fs(KERNEL_DS);
#ifdef KECHO
  while (sock_sendmsg(sock, &msg, (size_t)len) <= 0) {
#else
  while (sock_sendmsg(sock, &msg, (size_t)len) <= 0) {
#endif
    tries--;
    if (tries <= 0) break;
    set_fs(oldfs);
    schedule_timeout(2);
    set_fs(KERNEL_DS);
  }
  set_fs(oldfs);
  return len;
}

/* Socket send function. */
int my_sock_sendmsg(struct socket *sock, char *buf, int len, int deadline, int flags) 
{
  struct msghdr msg;
  struct iovec iov;
  mm_segment_t oldfs;
  int tries = TRIES;
  int flag = flags;

  if (len == 0)
    return 0;

  if (sock == NULL) 
    return 0;

  if (sock->sk == NULL) {
    printk("sk is null!\n");
    return 0;
  }

  if (flags != 1919)
    msg.msg_flags = MSG_DONTWAIT;
  else
    msg.msg_flags = 0;
  msg.msg_name = NULL;
  msg.msg_namelen = 0;
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;
  msg.msg_control = NULL;
  msg.msg_controllen = 0;
  iov.iov_base = (char *)buf;
  iov.iov_len = (__kernel_size_t)len;
  if (sock->type == SOCK_SEQPACKET)
    msg.msg_flags |= MSG_EOR;

  oldfs = get_fs();
  set_fs(KERNEL_DS);
  if (flag == 1919) {
    sock_sendmsg(sock, &msg, (size_t)len);
  } else {
#ifdef KECHO
    while (sock_sendmsg(sock, &msg, (size_t)len) <= 0) {
#else
    while (sock_sendmsg(sock, &msg, (size_t)len) <= 0) {
#endif
      tries--;
      if (tries <= 0) break;
      set_fs(oldfs);
      schedule_timeout(2);
      set_fs(KERNEL_DS);
    }
  }
  set_fs(oldfs);

  return len;
}

/* Socket send function. */
int my_sock_sendmsgv(struct socket *sock, struct iovec *iov, int len, int deadline,
	int flags) 
{
  struct msghdr msg;
  mm_segment_t oldfs;
  int tot_len = 0;
  int i;
  int tries = TRIES;

  int err;

  if (len == 0)
    return 0;

  if (sock == NULL) 
    return 0;

  if (sock->sk == NULL) {
    printk("sk is null!\n");
    return 0;
  }

  for (i = 0; i < len; i++)
    tot_len += iov[i].iov_len;

  if (flags != 1919)
    msg.msg_flags = MSG_DONTWAIT | MSG_NOSIGNAL;
  else
    msg.msg_flags = MSG_NOSIGNAL;
  msg.msg_name = NULL;
  msg.msg_namelen = 0;
  msg.msg_iov = iov;
  msg.msg_iovlen = len;
  msg.msg_control = NULL;
  msg.msg_controllen = 0;
  if (sock->type == SOCK_SEQPACKET)
    msg.msg_flags |= MSG_EOR;

  oldfs = get_fs();
  set_fs(KERNEL_DS);
  if (flags == 1919) {
    err = sock_sendmsg(sock, &msg, tot_len);
  } else {
#ifdef KECHO
    while ((err = sock_sendmsg(sock, &msg, tot_len)) <= 0) {
#else
    while ((err = sock_sendmsg(sock, &msg, tot_len)) <= 0) {
#endif
      tries--;
      if (tries <= 0) break;
      set_fs(oldfs);
      schedule_timeout(2);
      set_fs(KERNEL_DS);
    }
  }
  set_fs(oldfs);
  return err;
}

/* Socket UDP read function. */
int my_sock_recvfrom(struct socket *sock, char *buf, int len,
	unsigned flags, struct sockaddr *addr, int *addr_len, int sec) 
{
  struct msghdr msg;
  struct iovec iov;
  mm_segment_t oldfs;
  int ret = 0;
  char address[128];
  int tries = sec*HZ/2;

  if (len == 0)
    return 0;

  if (sock == NULL) 
    return 0;

  if (sock->sk == NULL) {
    printk("sk is null!\n");
    return 0;
  }

  msg.msg_flags = 0;
  msg.msg_name = address;
  msg.msg_namelen = 128;
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;
  msg.msg_control = NULL;
  msg.msg_controllen = 0;
  iov.iov_base = buf;
  iov.iov_len = len;

  if (tries <= 0) tries = 1;

  oldfs = get_fs();
  set_fs(KERNEL_DS);
  while (tries > 0) {
    ret = sock_recvmsg(sock, &msg, len, flags | MSG_DONTWAIT);
    if (ret > 0) {
      set_fs(oldfs);
      if (addr != NULL && msg.msg_namelen)
	memcpy(addr, address, msg.msg_namelen);
      return ret;
    }
    tries--;
    if (tries <= 0) {
      set_fs(oldfs);
      if (ret == 0 && addr != NULL && msg.msg_namelen)
	memcpy(addr, address, msg.msg_namelen);
      if (ret == -EAGAIN) return 0;
      else return ret;
    }
    set_fs(oldfs);
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(2);
    set_fs(KERNEL_DS);
  }
  set_fs(oldfs);

  if (ret == 0 && addr != NULL && msg.msg_namelen)
	memcpy(addr, address, msg.msg_namelen);

  if (ret == -EAGAIN) return 0;
  return ret;
}

/* Socket read function. */
int my_sock_recvmsg(struct socket *sock, char *buf, int len) 
{
  struct msghdr msg;
  struct iovec iov;
  mm_segment_t oldfs;
  int ret;

  if (len == 0)
    return 0;

  if (sock == NULL) 
    return 0;

  if (sock->sk == NULL) {
    printk("sk is null!\n");
    return 0;
  }

  msg.msg_flags = 0;
  msg.msg_name = NULL;
  msg.msg_namelen = 0;
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;
  msg.msg_control = NULL;
  msg.msg_controllen = 0;
  iov.iov_base = buf;
  iov.iov_len = len;

  oldfs = get_fs();
  set_fs(KERNEL_DS);
  ret = sock_recvmsg(sock, &msg, len, MSG_DONTWAIT);
  set_fs(oldfs);
  return ret;
}

/* Socket read function. */
int my_sock_recvmsgv(struct socket *sock, struct iovec *iov, int len) 
{
  struct msghdr msg;
  size_t tot_len = 0;
  int i;
  mm_segment_t oldfs;
  int ret;

  if (len == 0)
    return 0;

  if (sock == NULL) 
    return 0;

  if (sock->sk == NULL) {
    printk("sk is null!\n");
    return 0;
  }

  for (i = 0; i < len; i++)
    tot_len += iov[i].iov_len;

  msg.msg_flags = 0;
  msg.msg_name = NULL;
  msg.msg_namelen = 0;
  msg.msg_iov = iov;
  msg.msg_iovlen = len;
  msg.msg_control = NULL;
  msg.msg_controllen = 0;

  oldfs = get_fs();
  set_fs(KERNEL_DS);
  ret = sock_recvmsg(sock, &msg, tot_len, MSG_DONTWAIT);
  set_fs(oldfs);
  return ret;
}

